from enum import Enum
from typing import List


class WhereStatementKeys(Enum):
    EQUAL = 'EQUAL'
    NOEQUAL = 'NOEQUAL'
    MORE = 'MORE'
    LESS = 'LESS'
    MOREEQUAL = 'MOREEQUAL'
    LESSEQUAL = 'LESSEQUAL'


class WhereAttributesKeys(Enum):
    OPERATION = 'OPERATION'
    FIELD = 'FIELD'
    VALUE = 'VALUE'


where_stmts = {
    "==": WhereStatementKeys.EQUAL.value,
    "!=": WhereStatementKeys.NOEQUAL.value,
    "<": WhereStatementKeys.LESS.value,
    ">": WhereStatementKeys.MORE.value,
    "<=": WhereStatementKeys.LESSEQUAL.value,
    ">=": WhereStatementKeys.MOREEQUAL.value
}

where_stmts_reverse = {v: k for k, v in where_stmts.items()}


# FIXME: Need an AST for where conditions also check the order
def parse_where(where: str) -> List[dict]:
    if where is None or where == ():
        return []
    where_clauses = []
    w_clauses = where.split(",")
    for w in w_clauses:
        wh = w.split()

        if len(wh) == 3:
            try:
                stmt = {WhereAttributesKeys.OPERATION.value: where_stmts[wh[1]],
                        WhereAttributesKeys.FIELD.value: wh[0],
                        WhereAttributesKeys.VALUE.value: wh[2]}
            except KeyError:
                raise ValueError(f"Unknown \"where\" syntax.")
            where_clauses.append(stmt)
        else:
            raise ValueError(f"Unknown where syntax.")
    return where_clauses


def skip(row: dict, where: List[dict]) -> bool:
    if where is None or len(where) == 0:
        return False

    filter_wh = False
    for k in where:
        try:
            value = row[k[WhereAttributesKeys.FIELD.value]]
            data_value = f"\"{value}\"" if isinstance(value, str) and not value.isnumeric() else str(value)
            filter_wh = eval(data_value + ' ' +
                             str(where_stmts_reverse[k[WhereAttributesKeys.OPERATION.value]]) + ' ' +
                             str(k[WhereAttributesKeys.VALUE.value]))
            return not filter_wh
        except (KeyError, ValueError):
            return True
    return filter_wh



from os import cpu_count

import click as click

from openvariant.task.cat import cat as cat_task
from openvariant.task.count import count as count_task

from openvariant.task.groupby import group_by as group_by_task


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--debug', help='Show debug messages.', is_flag=True)
@click.version_option()
def cli(debug):
    # Configure the logging
    pass


@click.command(short_help='Concatenate files to standard input')
@click.argument('input_path', type=click.Path(exists=True), nargs=-1)
# @click.option('--columns', '-c', multiple=True, type=click.STRING, help="Extra columns to add")
@click.option('--where', '-w', multiple=True, type=click.STRING, default=None, help="Filter expression. ie: "
                                                                                    "CHROMOSOME == 4")
@click.option('--annotations', '-a', type=click.Path(exists=True))
def cat(input_path, where, annotations):
    cat_task(input_path[0], annotations, where)
    '''
    columns, global_annotations = __parse_global_annotations(columns, annotations)
    task_cat(
        __input_to_selection(input, global_annotations),
        columns=columns,
        where=__parse_where(where)
    )
    '''


@click.command(short_help='Count number of variants')
@click.argument('input_path', type=click.Path(exists=True), nargs=-1)
@click.option('--where', '-w', multiple=True, type=click.STRING)
@click.option('--groupby', '-g', type=click.STRING)
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
@click.option('--annotations', '-a', type=click.Path(exists=True))
def count(input_path, where, groupby, cores, quite, annotations):
    result = count_task(input_path[0], annotations, group_by=groupby, where=where, cores=cores, quite=quite)

    if len(result[1]) > 0:
        for k, v in sorted(result[1].items(), key=lambda v: v[1]):
            print("{}\t{}".format(k, v))

    print("TOTAL\t{}".format(result[0]))


@click.command(short_help='Groupby and run script')
@click.argument('input_files', type=click.Path(exists=True), nargs=1)
# @click.option('--script', '-s', type=click.STRING)
# @click.option('--columns', '-c', multiple=True, type=click.STRING, help="Extra columns to add")
@click.option('--where', '-w', multiple=True, type=click.STRING)
@click.option('--groupby', '-g', type=click.STRING)
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
@click.option('--annotations', '-a', type=click.Path(exists=True))
# @click.option('--headers', help='Send header as first row', is_flag=True)
def groupby(input_files, where, groupby, cores, quite, annotations):
    for group_key, r in group_by_task(input_files, annotations, key_by=groupby, where=where, cores=cores, quite=quite):
        print("{}".format(group_key))
        for line in r:
            print(' '.join(line))


cli.add_command(cat)
cli.add_command(count)
cli.add_command(groupby)

if __name__ == "__main__":
    cli()



import csv
from os import listdir
from os.path import isfile, join, isdir
import re
from typing import Generator, TextIO, List

from openvariant.annotation.annotation import Annotation
from openvariant.annotation.parser import AnnotationTypesParsers
from openvariant.config.config_annotation import AnnotationFormat, AnnotationGeneralKeys, ExcludesKeys, \
    AnnotationDelimiter
from openvariant.utils.logger import log


def _base_parser(lines: TextIO, delimiter: str) -> Generator[int, str, None]:
    try:
        read_tsv = csv.reader(lines, delimiter=AnnotationDelimiter[delimiter].value)
    except KeyError:
        raise KeyError(f"'{delimiter}' key not found.")
    for l_num, line in enumerate(read_tsv):  # lines, start=1):
        # Skip empty lines
        if len(line) == 0:
            continue

        # Skip comments
        if (line[0].startswith('#') or line[0].startswith('##') or line[0].startswith('browser') or line[0].startswith(
                'track')) \
                and not line[0].startswith('#CHROM'):
            continue

        yield l_num, line


def _parse_row(ann: dict, line: List, original_header: List, path: str, format_output: str) -> List[str]:
    annotations = ann[AnnotationGeneralKeys.ANNOTATION.name]
    row_parser = []
    for k, v in annotations.items():
        try:
            value = AnnotationTypesParsers[v[0]].value(v, line, original_header, path)
            row_parser.append(value)
        except IndexError:
            row_parser.append(float('nan'))

    return list(map(str, row_parser))


def _parser(file: str, annotation: dict, format_output: str, delimiter: str, display_header=True) -> Generator[List[str], None, None]:
    row = None
    fd = open(file, "rt")

    header = list(annotation[AnnotationGeneralKeys.ANNOTATION.name].keys())
    original_header = None
    for lnum, line in _base_parser(fd, delimiter):
        if original_header is None:
            original_header = line
            try:
                if not display_header:
                    continue
                row = header
            except (ValueError, KeyError) as e:
                log.error(f"Error parsing header {e}")
        else:
            try:
                row = _parse_row(annotation, line, original_header, file, format_output)
            except (ValueError, IndexError) as e:
                log.error(f"Error parsing line {lnum} {file} ({e, line, header})")
                continue

        yield row
    fd.close()


def _check_extension(ext: str, path: str) -> re.Match:
    rext = re.compile(ext[-1:] + "$")
    return rext.search(path)


def _extract_header(annotation: Annotation):
    return list(annotation.annotations.keys())


class Variant:

    def __init__(self, path: str, ann: Annotation) -> None:
        self._path: str = path
        self._annotation: Annotation = ann
        self._header: List[str] = _extract_header(ann)
        self._generator: Generator[List[str], None, None] = self._unify(path, ann)

    def _unify(self, base_path: str, annotation: Annotation, display_header=True) -> Generator[List[str], None, None]:
        an = annotation.structure
        format_output = annotation.format
        if isfile(base_path):
            for ext, ann in an.items():
                if _check_extension(ext, base_path):
                    for x in _parser(base_path, ann, format_output, annotation.delimiter, display_header):
                        display_header = False
                        yield x
        else:
            try:
                for file in listdir(base_path):
                    file_path = join(base_path, file)
                    if isfile(file_path):
                        for ext, ann in an.items():
                            if _check_extension(ext, file_path):
                                for x in _parser(file_path, ann, format_output, annotation.delimiter, display_header):
                                    display_header = False
                                    yield x
                    else:
                        for x in self._unify(file_path, annotation, display_header):
                            display_header = False
                            yield x
            except PermissionError as e:
                print(e)

    def _apply_exclude(self, line: List[str]) -> bool:
        for exclude in self._annotation.excludes:
            try:
                i = self._header.index(exclude[ExcludesKeys.FIELD.value])

                if str(exclude[ExcludesKeys.VALUE.value]) == line[i]:
                    return True
            except (KeyError, ValueError):
                return False
        return False

    def read(self) -> Generator[dict, None, None]:
        for i, line in enumerate(self._generator):
            if i != 0:
                if self._apply_exclude(line):
                    continue
                line_dict = {h: line[i] for i, h in enumerate(self._header)}
                yield line_dict

    def save(self, file_path: str, display_header=True):
        if isdir(file_path):
            raise ValueError("The path must be a file.")
        with open(file_path, "w") as file:
            writer = csv.writer(file, delimiter=AnnotationFormat[self._annotation.format.upper()].value)
            for i, line in enumerate(self._generator):
                if display_header and i == 0:
                    writer.writerow(line)
                elif i != 0:
                    if self._apply_exclude(line):
                        continue
                    writer.writerow(line)
            file.close()

    @property
    def path(self) -> str:
        return self._path

    @property
    def header(self) -> List[str]:
        return self._header

    @property
    def generator(self) -> Generator[List[str], None, None]:
        return self._generator

    @property
    def annotation(self) -> Annotation:
        return self._annotation



import glob
from fnmatch import fnmatch
from os import listdir
from os.path import isfile, join, basename
from typing import Generator
from copy import deepcopy

from openvariant.annotation.annotation import Annotation, merge_annotations_structure

from openvariant.config.config_annotation import ANNOTATION_EXTENSION, DEFAULT_RECURSIVE


def _get_annotation_file(annotation: Annotation, file_name: str, file_path: str, base_path: str) -> \
        Generator[str, Annotation, None]:
    ann = None
    pattern_matches = 0
    for pattern in annotation.patterns:
        if fnmatch(file_name, pattern):
            pattern_matches += 1
            ann = deepcopy(annotation)

    # Process filename and dirname annotations
    if ann is not None and pattern_matches > 0:  # and __where_match(file_annotations, where=where):
        yield file_path, ann


def find_files(base_path: str, annotation: Annotation) -> Generator[str, Annotation, None]:
    if annotation is None or base_path is None:
        raise ValueError('Annotation or path is missing')

    global_annotation = annotation

    if isfile(base_path):
        for f, a in _get_annotation_file(global_annotation, basename(base_path), base_path, base_path):
            yield f, a

    # Find files
    else:
        try:
            for file_name in listdir(base_path):
                file_path = join(base_path, file_name)
                if isfile(file_path):
                    local_annotation = global_annotation
                    for annotation_file in glob.iglob(join(base_path, "*.{}".format(ANNOTATION_EXTENSION))):
                        ann = Annotation(annotation_file)
                        recursive = DEFAULT_RECURSIVE if global_annotation.recursive is None \
                            else global_annotation.recursive
                        if recursive:
                            local_annotation = merge_annotations_structure(local_annotation, ann)
                    for f, a in find_files(file_path, local_annotation):
                        yield f, a
                else:
                    # Search subfolders
                    for f, a in find_files(file_path, global_annotation):
                        yield f, a
        except PermissionError as e:
            print(e)



from collections import defaultdict
from functools import partial
from multiprocessing import Pool
from os import cpu_count
from typing import Set, Generator, List, Tuple

from tqdm import tqdm

from openvariant.annotation.annotation import Annotation
from openvariant.task.find import find_files
from openvariant.utils.logger import log
from openvariant.utils.where import skip, parse_where
from openvariant.variant.variant import Variant


def _get_unique_values(file_path, annotation, key) -> Set:
    values = set()
    result = Variant(file_path, annotation)

    try:
        for r in result.read():
            values.add(r[key])
    except KeyError:
        log.warn(f"'{key}' key not found in '{file_path}' file")
    return values


def group(base_path: str, annotation: Annotation, key_by: str) -> Generator[str, List, None]:
    results = defaultdict(list)
    for file, ann in find_files(base_path, annotation):
        by_value = ann.annotations.get(key_by, None)

        if isinstance(by_value, tuple):
            values = _get_unique_values(file, ann, key_by)

            for s in values:
                a_clone = ann
                results[s].append((file, a_clone))

    for key, group_select in results.items():
        yield key, group_select


def group_by_task(selection, where=None) -> Tuple[str, List]:
    where_clauses = parse_where(where)
    group_key, group_values = selection
    output = []
    try:
        for value in group_values:
            result = Variant(value[0], value[1])
            for row in result.read():
                if skip(row, where_clauses):
                    continue
                output.append(list(row.values()))
    except BrokenPipeError:
        pass
    return group_key, output


def group_by(base_path: str, annotation_path: str, key_by: str, where=None, cores=cpu_count(), quite=False) -> \
        Generator[str, List, None]:
    annotation = Annotation(annotation_path)
    selection = list(group(base_path, annotation, key_by))

    with Pool(cores) as pool:
        task = partial(group_by_task, where=where)  # script=script, where=where_parsed, columns=columns, print_headers=headers)
        map_method = map if cores == 1 or len(selection) <= 1 else pool.imap_unordered

        for group_key, group_result in tqdm(
                map_method(task, selection),
                total=len(selection),
                desc="Computing groups".rjust(40),
                disable=(len(selection) < 2 or quite)
        ):
            yield group_key, group_result



from openvariant.annotation.annotation import Annotation
from openvariant.utils.format_line import format_line
from openvariant.utils.where import parse_where, skip
from openvariant.variant.variant import Variant


def cat(base_path: str, annotation_path: str, where=None) -> None:
    ann = Annotation(annotation_path)
    where_clauses = parse_where(where)

    result = Variant(base_path, ann)
    header = result.header
    print(format_line(header, result.annotation.format))
    for i, r in enumerate(result.read()):
        if skip(r, where_clauses):
            continue
        print(format_line(list(r.values()), result.annotation.format))


import functools
from multiprocessing import Pool
from os import cpu_count
from typing import Tuple, Union

from tqdm import tqdm

from openvariant.annotation.annotation import Annotation
from openvariant.task.find import find_files
from openvariant.utils.where import parse_where, skip
from openvariant.variant.variant import Variant


def _count_task(selection: Tuple[str, Annotation], group_by: str, where: str) -> Tuple[int, Union[dict, None]]:
    where_clauses = parse_where(where)
    i = 0
    input_file, input_annotations = selection
    result = Variant(input_file, input_annotations)

    header = result.header
    if group_by is None:
        for r in result.read():
            if skip(r, where_clauses):
                continue
            i += 1
        return i, None
    else:
        groups = {}
        # key_not_found = False
        for r in result.read():
            if skip(r, where_clauses):
                continue
            try:
                val = list(r.values())[header.index(group_by)]
                val_count = groups.get(val, 0)
                groups[val] = val_count + 1
                i += 1
            except ValueError:
                pass

        return i, groups


def count(base_path: str, annotation_path: str, group_by=None, where=None, cores=cpu_count(), quite=False) -> \
        Tuple[int, Union[None, dict]]:
    ann = Annotation(annotation_path)

    selection = []
    for k, a in find_files(base_path, ann):
        selection += [(k, a)]

    with Pool(cores) as pool:
        groups = {}
        task = functools.partial(_count_task, group_by=group_by, where=where[0])
        map_method = pool.imap_unordered if len(selection) > 1 else map
        total = 0
        for c, g in tqdm(map_method(task, selection),
                         total=len(selection),
                         desc="Counting variants".rjust(40),
                         disable=(len(selection) < 2) or quite
                         ):

            # Update groups
            if g is not None:
                for k, v in g.items():
                    val = groups.get(k, 0)
                    groups[k] = val + v
            total += c

    return total, groups



from enum import Enum
from functools import partial
from os.path import basename, dirname
from typing import Tuple, Any, List, Optional, Union, Callable

from openvariant.annotation.builder import Builder


def _get_text_by_field(x: List, line: List, original_header: List, func: Builder or None):
    value = None
    for y in x:
        value = line[original_header.index(y)] if y in original_header else None
        if value is not None:
            value = func(value) if func is not None else value
        if value is not None:
            break
    if value is None:
        return float('nan')
    return value


def _static_parser(x: Tuple[str, Any], line: List, original_header: List, path: str) -> str:
    return x[1]


def _internal_parser(x: Tuple[str, List, Builder], line: List, original_header: List, path: str) -> \
        Optional[Union[int, str, float]]:
    #print(x[1], line, original_header, x[2])
    return _get_text_by_field(x[1], line, original_header, x[2])


def _filename_parser(x: Tuple[str, Builder], line: List, original_header: List, path: str) -> str:
    value = x[1](basename(path))
    return value if value is not None else float('nan')


def _dirname_parser(x: Tuple[str, Builder], line: List, original_header: List, path: str) -> str:
    value = x[1](basename(dirname(path)))
    return value if value is not None else float('nan')


def _plugin_parser(x: Tuple[str, List, Callable], line: List, original_header: List, path: str) -> str:
    value = _get_text_by_field(x[1], line, original_header, None)
    value = x[2](value)
    return value if value is not None else float('nan')


class AnnotationTypesParsers(Enum):
    STATIC = partial(_static_parser)
    INTERNAL = partial(_internal_parser)
    FILENAME = partial(_filename_parser)
    DIRNAME = partial(_dirname_parser)
    PLUGIN = partial(_plugin_parser)
    '''
    MAPPING = partial(_mapping_builder)
    '''



import importlib
from enum import Enum
from functools import partial
from typing import List, Tuple, Callable, Any

from openvariant.config.config_annotation import AnnotationTypes, AnnotationKeys


class Builder:
    func: str = None

    def __init__(self, func: str) -> None:
        self.func = func

    def __call__(self, x: Any) -> Any:
        return eval(self.func)(x)


def _static_builder(x: dict) -> Tuple[str, Any]:
    return AnnotationTypes.STATIC.name, x[AnnotationKeys.VALUE.value]


def _internal_builder(x: dict) -> Tuple[str, List, Builder]:
    return AnnotationTypes.INTERNAL.name, x[AnnotationKeys.FIELD_SOURCE.value], Builder("(lambda y: y)") \
        if AnnotationKeys.FUNCTION.value not in x or len(x[AnnotationKeys.FUNCTION.value]) == 2 \
        else Builder(x[AnnotationKeys.FUNCTION.value])


def _dirname_builder(x: dict) -> Tuple[str, Builder]:
    return AnnotationTypes.DIRNAME.name, Builder("(lambda y: y)") \
        if x[AnnotationKeys.FUNCTION.value] is None or len(x[AnnotationKeys.FUNCTION.value]) == 2 \
        else Builder(x[AnnotationKeys.FUNCTION.value])


def _filename_builder(x: dict) -> Tuple[str, Builder]:
    return AnnotationTypes.FILENAME.name, Builder("(lambda y: y)") \
        if AnnotationKeys.FUNCTION.value not in x or len(x[AnnotationKeys.FUNCTION.value]) == 2 \
        else Builder(x[AnnotationKeys.FUNCTION.value])


def _plugin_builder(x: dict) -> Tuple[str, List, Callable]:
    try:
        mod = importlib.import_module(f".{x[AnnotationTypes.PLUGIN.value]}", package="plugins")
        func = getattr(mod, x[AnnotationTypes.PLUGIN.value])
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"Enable to found 'plugins.{x[AnnotationTypes.PLUGIN.value]}' module.")
    return AnnotationTypes.PLUGIN.name, x[AnnotationKeys.FIELD_SOURCE.value], func


'''
def _mapping_builder(x: dict) -> Tuple[str, List]:
    values: List = []
    with open(x[AnnotationKeys.MAPPING_FILE.value]) as f:
        for r in csv.DictReader(f, delimiter='\t'):
            val = r[x[AnnotationKeys.MAPPING_FIELD.value]]
            values.append(val)
    return AnnotationTypes.MAPPING.name, values
'''


class AnnotationTypesBuilders(Enum):
    STATIC = partial(_static_builder)
    INTERNAL = partial(_internal_builder)
    DIRNAME = partial(_dirname_builder)
    FILENAME = partial(_filename_builder)
    PLUGIN = partial(_plugin_builder)
    # MAPPING = partial(_mapping_builder)



from setuptools import setup, find_packages

setup(
    name="open-variant",
    version="0.0.1",
    author="BBGLab - Barcelona Biomedical Genomics Lab",
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests', 'plugins']),
    include_package_data=True,
    install_requires=['pyyaml', 'tqdm', 'click'],
    entry_points={
        'console_scripts': [
            'openvar = openvariant.task.openvar:cli',
        ]
    },
)
